#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "pssp.h"



struct atom wksp[NATINMOL];
int lastatom; // last atom loaded into atoms[]

void putatoms(struct descr *d){	/* put the atoms into the cell list according to params */
  int cmdline, mol, ninmol, i;
  lastatom = 0;

// loop through lines stored which define the structure
  for(cmdline = 0; putcmd[cmdline][0] != '\0'; cmdline++){
    // printf("putatoms line %d is %s\n", cmdline, putcmd[cmdline]);
   
    if(strncmp(putcmd[cmdline], "getmol", 6) == 0){
      mol = atoi ( putcmd[cmdline] + 7 );
      ninmol = molecule[mol].m_natoms;
      if(ninmol > NATINMOL) errout("RECOMPILE -- NATINMOL too small");
      // printf("getmol mol = %d, ninmol = %d\n", mol, ninmol);
      for(i=0; i<ninmol; i++){
	atcopy(&molecule[mol].m_atom[i], &wksp[i]);
	// printf("atcopy %f %f %f\n", wksp[i].a_xyz[0], wksp[i].a_xyz[1], wksp[i].a_xyz[2]);
      }
    }

    else if(strncmp(putcmd[cmdline], "rot_axis", 8) == 0){ // rotate some atoms in wksp
      rot_axis(putcmd[cmdline]+9, d);
    }

    else if(strncmp(putcmd[cmdline], "rot_body_fix", 12) == 0){
      rot_body_fix(putcmd[cmdline]+13, ninmol);
    }

    else if(strncmp(putcmd[cmdline], "rot_body_var", 12) == 0){
      rot_body_var(putcmd[cmdline]+13, ninmol, d);
    }

    else if(strncmp(putcmd[cmdline], "putmol", 6) == 0){ // put current molecule into cell list
      putmol(putcmd[cmdline]+7, ninmol, d);
    }

    /* ********  include following printf statements for debugging */
    //printf("wksp holds:\n");
    //    for(i=0; i<ninmol; i++) printf("%3d %8.3f %8.3f %8.3f\n", wksp[i].a_atnum, wksp[i].a_xyz[0], wksp[i].a_xyz[1], wksp[i].a_xyz[2]) ;
  }
}

// rotate selected atoms in molecule in the workspace about selected axis
void rot_axis(char *cmd, struct descr *d){
   // argument is a string consisting of 2 atoms defining axis, then parameter
   // no., then list of atoms to rotate.
   // positive direction is defined by RH rule for vector from atom1 to atom2

   float a[3][3], b[3], c[3], sa, ca, tmp;
   int atom1, atom2, i, j, k, atnum, rotpar;
   char *cp;

   // printf("In rot_axis.\n");
   init_pluck(cmd);
   atom1 = atoi( pluck() );
   atom2 = atoi( pluck() );

   // printf("atom1 = %d, atom2 = %d\n ", atom1, atom2); 
   atom1--;
   atom2--;
   rotpar = atoi( pluck() ) - 1;      // parameters numbered from 1 in input file
   ca = cos(d->d_f[rotpar] / RADIAN);
   sa = sin(d->d_f[rotpar] / RADIAN);

   for(i=0; i<3; i++) a[0][i] = wksp[atom2].a_xyz[i] - wksp[atom1].a_xyz[i];
   if( norm_vec(a[0]) == 0) errout("atom1=atom2 in rot_axis_var.");
   setup_rot(a);
   // now a[][] defines a rotation that puts the chosen axis along x
   /*   for(i=0; i<3; i++){ 
     for(j=0; j<3; j++) 
       printf("%8.4f ", a[i][j]); 
       printf("\n");}// print matrix a */

   while(1){
     cp = pluck();
     if(*cp == '\0') break;
     atnum = atoi(cp)-1; // atoms numbered externally form 1, internally from 0
     // printf(" %d ", atnum);
     for(j=0; j<3; j++) b[j] = wksp[atnum].a_xyz[j] - wksp[atom1].a_xyz[j];
     for(j=0; j<3; j++){
      	c[j] = 0;
      	for(k=0; k<3; k++) c[j] += a[j][k] * b[k];
     }
     tmp  = ca * c[1] - sa * c[2]; // rotate c[] about x axis
     c[2] = ca * c[2] + sa * c[1];
     c[1] = tmp;
     for(j=0; j<3; j++){      // transform back, making use of the fact that a^-1 = a^T
       b[j] = 0;
       for(k=0; k<3; k++) b[j] += a[k][j] * c[k];
     }
     for(j=0; j<3; j++) wksp[atnum].a_xyz[j] =b[j] + wksp[atom1].a_xyz[j];
   }
   // printf("\nLeaving rot_axis.\n");
}

void setup_rot(float a[3][3]){
// given first row of a, creates 2 others as perpendicular vectors.
   int i, k;
   float b[3], dot;

   for(i=0; i<3; i++) a[1][i] = 0;
   /* choose b, which had better not be parallel to a[0] */
   if( fabs (a[0][0] < 0.7)) k=0; // b = (1,0,0), AdotB = a[0][0];
   else k=1;                      // b = (0,1,0)
   dot = 1. / sqrt(1. - a[0][k] * a[0][k]);
   a[1][k] = dot;
   dot *= -a[0][k];
   for(i=0; i<3; i++) a[1][i] += dot * a[0][i]; // now a[1] is perp to a[0]
   for(i=0; i<3; i++) 
     a[2][i] = a[0][(i+1)%3] * a[1][(i+2)%3] - a[0][(i+2)%3] * a[1][(i+1)%3];
     // a[2] = a[0] cross a[1]
}

int norm_vec(float a[3]){
   float z=0;
   int i;

   for(i=0; i<3; i++) z += a[i]*a[i];
   if(z < 1.0e-10) return(0); // error
   z = sqrt(z);
   for(i=0; i<3; i++) a[i] /= z;
   return(1);
}

// rotate molecule in workspace as a rigid body about one cart. axis
// the plan is for these conventions to be identical to GSAS
void rot_body_fix(char *cmd, int ninm){
  // arguments are direction [xyz] and angle in degrees
  float a[3][3], b[3], ang;
  int i, j, axis, ax1, ax2;
  char *cp, *ap;
  
  for(i=0; i<3; i++) for(j=0; j<3; j++) a[i][j] = 0;
  
  init_pluck(cmd);
  cp=pluck();
  ap = pluck();
  if(*ap == '\0' || *cp == '\0') errout("in rot_body_fix, no params");
  
  ang = atof(ap) / RADIAN;
  
  axis = *cp - 'x';
  if(axis < 0 || axis > 2) axis = *cp - 'X';
  ax1 = (axis+1)%3;
  ax2 = (axis+2)%3;
  if(axis < 0 || axis > 2) errout("in rot_body_fix, cant understand axis");
  a[axis][axis] = 1;
  a[ax1][ax1] = a[ax2][ax2] = cos(ang);
  a[ax2][ax1] = -1 * ( a[ax1][ax2] = sin(ang) );
  /*  printf("In rot_body_fix, rotation matrix:\n");
  for(i=0; i<3; i++){ 
    for(j=0; j<3; j++) 
      printf("%8.4f ", a[i][j]); 
    printf("\n");
    }// print matrix a */

  // now use a[][] to rotate all molecules in wksp
  while(ninm--){ // count down through all atoms in list
    for(i=0; i<3; i++){
      b[i] = 0;
      for(j=0; j<3; j++) b[i] += a[i][j] * wksp[ninm].a_xyz[j];
    }
    for(i=0; i<3; i++) wksp[ninm].a_xyz[i] = b[i];
  }
}

void rot_body_var(char *cmd, int ninm, struct descr *d){
// rotate molecule in workspace as a rigid body about one cart. axis
// the plan is for these conventions to be identical to GSAS
// arguments are direction [xyz] and parameter number specifying angle in degrees
  float a[3][3], b[3], ang;
  int i, j, axis, ax1, ax2, pn;
  char *cp, *ap;
  
  for(i=0; i<3; i++) for(j=0; j<3; j++) a[i][j] = 0;
  
  init_pluck(cmd);
  cp=pluck();
  ap = pluck();
  if(*ap == '\0' || *cp == '\0') errout("in rot_body_var, no params");
  ang = d->d_f[atoi(ap) - 1] / RADIAN; // parameters start from 1 in input file
  
  axis = *cp - 'x';
  if(axis < 0 || axis > 2) axis = *cp - 'X';
  ax1 = (axis+1)%3;
  ax2 = (axis+2)%3;
  if(axis < 0 || axis > 2) errout("in rot_body_fix, cant understand axis");
  a[axis][axis] = 1;
  a[ax1][ax1] = a[ax2][ax2] = cos(ang);
  a[ax2][ax1] = -1 * ( a[ax1][ax2] = sin(ang) );
  /* printf("In rot_body_var, rotation matrix:\n");
  for(i=0; i<3; i++){ 
    for(j=0; j<3; j++) 
      printf("%8.4f ", a[i][j]); 
    printf("\n");
    }// print matrix a */

  // now use a[][] to rotate all molecules in wksp
  while(ninm--){ // count down through all atoms in list
    for(i=0; i<3; i++){
      b[i] = 0;
      for(j=0; j<3; j++) b[i] += a[i][j] * wksp[ninm].a_xyz[j];
    }
    for(i=0; i<3; i++) wksp[ninm].a_xyz[i] = b[i];
  }
}



// put molecule from workspace into atoms[], transforming from cartesian
// to fractional coords, starting at lastatom.
// arguments are crystal x y z coordinates of origin, and parameter numbers for
// translations about each of the crystal axes.  Neg. param OK.
void putmol(char *cmd, int ninm, struct descr *d){
  float txyz[3], a[3], b[3];
  int pxyz[3];
  struct symelement *sp;
  int  i, j, k, isym;

  init_pluck(cmd);
  txyz[0] = atof( pluck() ); // no error checking isn't good
  txyz[1] = atof( pluck() );
  txyz[2] = atof( pluck() );
  pxyz[0] = atoi( pluck() );
  pxyz[1] = atoi( pluck() );
  pxyz[2] = atoi( pluck() );

  /*   printf("%s\n" , cmd);
printf("%f\n", txyz[0]);
printf("%f\n", txyz[1]);
printf("%f\n", txyz[2]);
printf("%d\n", pxyz[0]);
printf("%d\n", pxyz[1]);
printf("%d\n", pxyz[2]); */

  for(i=0; i<ninm; i++){
    atcopy(&wksp[i], &atoms[i+lastatom]); // copy atom name etc.
    for(j=0; j<3; j++){
      b[j] = txyz[j]; // load origin, both constant and variable params
      if(pxyz[j] > 0) b[j] += d->d_f[pxyz[j] -1 ];  // parameters in input file start from 1
      else if(pxyz[j] < 0) b[j] -= d->d_f[-pxyz[j] -1 ];
      // otherwise, pxyz[j] == 0, and no parameter adds to position
      
      for(k=0; k<3; k++) b[j] += cellconv[j][k] * wksp[i].a_xyz[k];
      atoms[i+lastatom].a_xyz[j] = b[j];
    }
  } // end of loop on i = number of atom in this molecule
  lastatom += ninm;
  if(lastatom > NATS) errout("RECOMPILE -- NATS too small");

}

void atcopy(struct atom *p1, struct atom *p2){ // copy atom from p1 to p2
  int i;
  char *c1, *c2;
  c1 = (char*) p1;
  c2 = (char*) p2;
  // printf(",");
  for(i=0; i<sizeof(struct atom); i++, c1++, c2++) *c2 = *c1;
  // printf(".");
}

char pluckbuf[MAXBUF];
char *sp, *ep, *eep; // pointers to start and end of next string in pluck, and end of s_buf.

/* copy argument to pluckbuf so that elements can be plucked one by one */
void init_pluck(char *ap){
  eep = ep = sp = pluckbuf;

  while(eep < pluckbuf + MAXBUF){
    *eep = *ap;
    if(*eep == '\0')
      return;
    ap++;
    eep++;
  }
  errout("overflow loading pluckbuf.");
}

// pluck returns a pointer to the next variable in pluckbuf after null-terminating it.
char *pluck(){  
  if(ep == eep)   // have already returned last argument, now terminate
    return ep;    // pointer to '\0' is signal of end of param list

  if(ep == sp)
    ep = sp-1;

  ep++;

  while( *ep == ' ') ep++;
  sp = ep;
  while( *ep != ' '&& *ep != '\0' ) ep++;
  *ep = '\0';     // null-terminate string of this element

  return sp;
}































